import asyncio
import fnmatch
import importlib
import logging
import os
import sys
from dataclasses import dataclass, field
from threading import Lock
from typing import Dict, List, Optional, Union

from braintrust_core.util import eprint

from .. import login
from ..framework import (
    Evaluator,
    EvaluatorInstance,
    ReporterDef,
    _evals,
    _set_lazy_load,
    bcolors,
    default_reporter,
    init_experiment,
    parse_filters,
    report_evaluator_result,
    run_evaluator,
    set_thread_pool_max_workers,
)

INCLUDE = [
    "**/eval_*.py",
]
EXCLUDE = ["**/site-packages/**"]

_logger = logging.getLogger("braintrust.eval")


_import_lock = Lock()


@dataclass
class FileHandle:
    in_file: str

    def rebuild(self):
        in_file = os.path.abspath(self.in_file)

        with _import_lock:
            with _set_lazy_load(True):
                _evals.clear()

                try:
                    # https://stackoverflow.com/questions/67631/how-can-i-import-a-module-dynamically-given-the-full-path
                    spec = importlib.util.spec_from_file_location("eval", in_file)
                    module = importlib.util.module_from_spec(spec)
                    spec.loader.exec_module(module)

                    ret = _evals.copy()
                finally:
                    _evals.clear()

        return ret

    def watch(self):
        raise NotImplementedError


@dataclass
class EvaluatorOpts:
    verbose: bool
    no_send_logs: bool
    no_progress_bars: bool
    terminate_on_failure: bool
    watch: bool
    filters: List[str]
    jsonl: bool


@dataclass
class LoadedEvaluator:
    handle: FileHandle
    evaluator: Evaluator
    reporter: Optional[Union[ReporterDef, str]] = None


@dataclass
class EvaluatorState:
    evaluators: Dict[str, LoadedEvaluator] = field(default_factory=dict)
    reporters: Dict[str, ReporterDef] = field(default_factory=dict)


def update_evaluators(evaluators: EvaluatorState, handles, terminate_on_failure):
    for handle in handles:
        try:
            module_evals = handle.rebuild()
        except Exception as e:
            if terminate_on_failure:
                raise
            else:
                eprint(f"Failed to import {handle.in_file}: {e}")
                continue

        for eval_name, evaluator in module_evals.evaluators.items():
            if not isinstance(evaluator, EvaluatorInstance):
                continue

            if eval_name in evaluators.evaluators:
                _logger.warning(
                    f"Evaluator {eval_name} already exists (in {evaluators[eval_name].handle.in_file} and {handle.in_file}). Will skip {eval_name} in {handle.in_file}."
                )
                continue

            evaluators.evaluators[eval_name] = LoadedEvaluator(
                handle=handle, evaluator=evaluator.evaluator, reporter=evaluator.reporter
            )

        for reporter_name, reporter in module_evals.reporters.items():
            if not isinstance(reporter, ReporterDef):
                continue

            if reporter_name in evaluators.reporters:
                _logger.warning(
                    f"Reporter {reporter_name} already exists (in {evaluators.reporters[reporter_name].module} and {handle.in_file}). Will skip {reporter_name} in {handle.in_file}."
                )
                continue

            evaluators.reporters[reporter_name] = reporter


async def run_evaluator_task(evaluator, position, opts: EvaluatorOpts):
    experiment = None
    if not opts.no_send_logs:
        experiment = init_experiment(
            evaluator.project_name,
            evaluator.experiment_name,
            metadata=evaluator.metadata,
            is_public=evaluator.is_public,
        )

    try:
        return await run_evaluator(
            experiment, evaluator, position if not opts.no_progress_bars else None, opts.filters
        )
    finally:
        if experiment:
            experiment.flush()


def resolve_reporter(reporter: Optional[Union[ReporterDef, str]], reporters: Dict[str, ReporterDef]) -> ReporterDef:
    if isinstance(reporter, str):
        if reporter not in reporters:
            raise ValueError(f"Reporter {reporter} not found")
        return reporters[reporter]
    elif reporter:
        return reporter
    elif not reporters:
        return default_reporter
    elif len(reporters) == 1:
        return next(iter(reporters.values()))
    else:
        reporter_names = ", ".join(reporters.keys())
        raise ValueError(f"Multiple reporters found ({reporter_names}). Please specify a reporter explicitly.")


def add_report(eval_reports, reporter, report):
    if reporter.name not in eval_reports:
        eval_reports[reporter.name] = {"reporter": reporter, "results": []}
    eval_reports[reporter.name]["results"].append(report)


async def run_once(handles, evaluator_opts):
    objects = EvaluatorState()
    update_evaluators(objects, handles, terminate_on_failure=evaluator_opts.terminate_on_failure)

    eval_promises = [
        asyncio.create_task(run_evaluator_task(evaluator.evaluator, idx, evaluator_opts))
        for idx, evaluator in enumerate(objects.evaluators.values())
    ]
    eval_results = [await p for p in eval_promises]

    eval_reports = {}
    for evaluator, result in zip(objects.evaluators.values(), eval_results):
        resolved_reporter = resolve_reporter(evaluator.reporter, objects.reporters)
        report = resolved_reporter._call_report_eval(
            evaluator=evaluator.evaluator, result=result, verbose=evaluator_opts.verbose, jsonl=evaluator_opts.jsonl
        )
        add_report(eval_reports, resolved_reporter, report)

    all_success = True
    for report in eval_reports.values():
        reporter = report["reporter"]
        results = [await r for r in report["results"]]
        if not await reporter._call_report_run(results, verbose=evaluator_opts.verbose, jsonl=evaluator_opts.jsonl):
            _logger.error(f"Reporter {reporter.name} failed")
            all_success = False

    return all_success


def check_match(path_input, include_patterns, exclude_patterns):
    p = os.path.abspath(path_input)
    if include_patterns:
        include = False
        for pattern in include_patterns:
            if fnmatch.fnmatch(p, pattern):
                include = True
                break
        if not include:
            return False

    if exclude_patterns:
        exclude = False
        for pattern in exclude_patterns:
            if fnmatch.fnmatch(p, pattern):
                exclude = True
                break
        return not exclude

    return True


def collect_files(input_path):
    if os.path.isdir(input_path):
        for root, dirs, files in os.walk(input_path):
            for file in files:
                fname = os.path.join(root, file)
                if check_match(fname, INCLUDE, EXCLUDE):
                    yield fname
    else:
        if not check_match(input_path, INCLUDE, EXCLUDE):
            _logger.warning(
                f"Reading {input_path} because it was specified directly. Rename it to eval_*.py "
                + "to include it automatically when you specify a directory."
            )
        yield input_path


def initialize_handles(files):
    input_paths = files if len(files) > 0 else ["."]

    fnames = set()
    for path in input_paths:
        for fname in collect_files(path):
            fnames.add(os.path.abspath(fname))

    return [FileHandle(in_file=fname) for fname in fnames]


def run(args):
    if args.num_workers:
        set_thread_pool_max_workers(args.num_workers)

    evaluator_opts = EvaluatorOpts(
        verbose=args.verbose,
        no_send_logs=args.no_send_logs,
        no_progress_bars=args.no_progress_bars,
        terminate_on_failure=args.terminate_on_failure,
        watch=args.watch,
        filters=parse_filters(args.filter) if args.filter else [],
        jsonl=args.jsonl,
    )

    handles = initialize_handles(args.files)

    if not evaluator_opts.no_send_logs:
        login(
            api_key=args.api_key,
            org_name=args.org_name,
            app_url=args.app_url,
        )

    if args.watch:
        eprint("Watch mode is not yet implemented")
        exit(1)
    else:
        if not asyncio.run(run_once(handles, evaluator_opts)):
            sys.exit(1)


def build_parser(subparsers, parent_parser):
    parser = subparsers.add_parser(
        "eval",
        help="Run evals locally.",
        parents=[parent_parser],
    )

    parser.add_argument(
        "--api-key",
        help="Specify a braintrust api key. If the parameter is not specified, the BRAINTRUST_API_KEY environment variable will be used.",
    )
    parser.add_argument(
        "--org-name",
        help="The name of a specific organization to connect to. This is useful if you belong to multiple.",
    )
    parser.add_argument(
        "--app-url",
        help="Specify a custom braintrust app url. Defaults to https://www.braintrustdata.com. This is only necessary if you are using an experimental version of Braintrust",
    )
    parser.add_argument(
        "--watch",
        action="store_true",
        help="Watch files for changes and rerun evals when changes are detected",
    )
    parser.add_argument(
        "--filter",
        help="Only run evaluators that match these filters. Each filter is a regular expression (https://docs.python.org/3/library/re.html). For example, --filter metadata.priority='^P0$' input.name='foo.*bar' will only run evaluators that have metadata.priority equal to 'P0' and input.name matching the regular expression 'foo.*bar'.",
        nargs="*",
    )
    parser.add_argument(
        "--jsonl",
        help="Format score summaries as jsonl, i.e. one JSON-formatted line per summary.",
        action="store_true",
    )
    parser.add_argument(
        "--no-send-logs",
        action="store_true",
        help="Do not send logs to Braintrust. Useful for testing evaluators without uploading results.",
    )
    parser.add_argument(
        "--no-progress-bars",
        action="store_true",
        help="Do not show progress bars when processing evaluators.",
    )
    parser.add_argument(
        "--terminate-on-failure",
        action="store_true",
        help="If provided, terminates on a failing eval, instead of the default (moving onto the next one).",
    )
    parser.add_argument(
        "--num-workers",
        type=int,
        help="Specify the number of concurrent worker threads to run evals over, if they are defined as synchronous functions. Async functions will be run in the single-threaded asyncio event loop. If not specified, defaults to the number of cores on the machine.",
    )
    parser.add_argument(
        "files",
        nargs="*",
        help="A list of files or directories to run. If no files are specified, the current directory is used.",
    )

    parser.set_defaults(func=run)



import argparse
import textwrap

_module_not_found_error = None
try:
    from . import api, logs, redshift
except ModuleNotFoundError as e:
    _module_not_found_error = e


def fail_with_module_not_found_error(*args, **kwargs):
    raise ModuleNotFoundError(
        textwrap.dedent(
            f"""\
            At least one dependency not found: {str(_module_not_found_error)!r}
            It is possible that braintrust was installed without the CLI dependencies. Run:

              pip install 'braintrust[cli]'

            to install braintrust with the CLI dependencies (make sure to quote 'braintrust[cli]')."""
        )
    )


def build_parser(subparsers, parent_parser):
    install_parser = subparsers.add_parser(
        "install",
        help="Tools to setup and verify Braintrust's installation in your environment.",
        parents=[parent_parser],
    )
    if _module_not_found_error:
        install_parser.add_argument("args", nargs=argparse.REMAINDER)
        install_parser.set_defaults(func=fail_with_module_not_found_error)
    else:
        install_subparsers = install_parser.add_subparsers(dest="install_subcommand", required=True)

        for module in [api, logs, redshift]:
            module.build_parser(install_subparsers, parents=[parent_parser])



import inspect
import os.path
import sys
import threading
import urllib.parse
from typing import Any, Callable, Generic, TypeVar

from requests import HTTPError

GLOBAL_PROJECT = "Global"


class AugmentedHTTPError(Exception):
    pass


def response_raise_for_status(resp):
    try:
        resp.raise_for_status()
    except HTTPError as e:
        raise AugmentedHTTPError(f"{resp.text}") from e


def get_caller_location():
    # Modified from
    # https://stackoverflow.com/questions/24438976/debugging-get-filename-and-line-number-from-which-a-function-is-called
    # to fetch the first stack frame not contained inside the same directory as
    # this file.
    this_dir = None
    call_stack = inspect.stack()
    for frame in call_stack:
        caller = inspect.getframeinfo(frame.frame)
        if this_dir is None:
            this_dir = os.path.dirname(caller.filename)
        if os.path.dirname(caller.filename) != this_dir:
            return dict(
                caller_functionname=caller.function,
                caller_filename=caller.filename,
                caller_lineno=caller.lineno,
            )
    return None


T = TypeVar("T")


class LazyValue(Generic[T]):
    """A simple wrapper around a callable object which computes the value
    on-demand and saves it for future retrievals.
    """

    def __init__(self, callable: Callable[[], T], use_mutex: bool):
        self.callable = callable
        self.mutex = threading.Lock() if use_mutex else None
        self.has_computed = False
        self.value = None

    def get(self) -> T:
        # Short-circuit check `has_computed`. This should be fine because
        # setting `has_computed` is atomic and python should have sequentially
        # consistent semantics, so we'll observe the write to `self.value` as
        # well.
        if self.has_computed:
            return self.value
        if self.mutex:
            self.mutex.acquire()
        try:
            if not self.has_computed:
                res = self.callable()
                self.value = res
                self.has_computed = True
            return self.value
        finally:
            if self.mutex:
                self.mutex.release()



import logging
import textwrap
import time

from botocore.exceptions import ClientError

from ...aws import cloudformation

_logger = logging.getLogger("braintrust.install.api")

PARAMS = {
    "OrgName": "org_name",
    "ProvisionedConcurrency": "provisioned_concurrency",
    "EncryptDatabase": "encrypt_database",
    "PostgresAlternativeHost": "postgres_alternative_host",
    "APIHandlerMemorySize": "api_handler_memory_size",
    "WhitelistedOrigins": "whitelisted_origins",
    "PublicSubnet1AZ": "public_subnet_1_az",
    "PrivateSubnet1AZ": "private_subnet_1_az",
    "PrivateSubnet2AZ": "private_subnet_2_az",
    "PrivateSubnet3AZ": "private_subnet_3_az",
    "VPCCIDR": "vpc_cidr",
    "PublicSubnet1CIDR": "public_subnet_1_cidr",
    "PrivateSubnet1CIDR": "private_subnet_1_cidr",
    "PrivateSubnet2CIDR": "private_subnet_2_cidr",
    "PrivateSubnet3CIDR": "private_subnet_3_cidr",
    "ManagedClickhouse": "managed_clickhouse",
    "ClickhouseInstanceType": "clickhouse_instance_type",
}

REMOVED_PARAMS = ["ThirdAZIndex"]

DEFAULTS = {
    "DwType": "Postgres",
    "EncryptDatabase": "false",
    "ProvisionedConcurrency": 0,
    "APIHandlerMemorySize": 10240,
}

CAPABILITIES = ["CAPABILITY_IAM", "CAPABILITY_AUTO_EXPAND"]

LATEST_TEMPLATE = "https://braintrust-cf.s3.amazonaws.com/braintrust-latest.yaml"


def build_parser(subparsers, parents):
    parser = subparsers.add_parser("api", help="Install the Braintrust function API", parents=parents)

    parser.add_argument("name", help="Name of the CloudFormation stack to create or update")
    parser.add_argument(
        "--create",
        help="Create the stack if it does not exist",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--vpc-connect",
        help="Connect to an existing VPC",
        action="store_true",
        default=False,
    )
    parser.add_argument(
        "--template",
        help="A specific CloudFormation template URL to use",
        default=None,
    )
    parser.add_argument(
        "--update-template",
        help="Update the CloudFormation to the latest version of the template",
        action="store_true",
        default=False,
    )

    # OrgName, ProvisionedConcurrency
    parser.add_argument("--org-name", help="The name of your organization", default=None)
    parser.add_argument(
        "--provisioned-concurrency",
        help="The amount of provisioned concurrency",
        default=None,
        type=int,
    )
    parser.add_argument(
        "--api-handler-memory-size",
        help="The amount of memory to allocate to the API handler",
        default=None,
        type=int,
    )
    parser.add_argument(
        "--whitelisted-origins",
        help="Comma-separated list of origins to whitelist",
        default=None,
    )
    parser.add_argument(
        "--public-subnet-1-az",
        help="The availability zone for the public subnet",
        default=None,
    )
    parser.add_argument(
        "--private-subnet-1-az",
        help="The availability zone for private subnet 1",
        default=None,
    )
    parser.add_argument(
        "--private-subnet-2-az",
        help="The availability zone for private subnet 2",
        default=None,
    )
    parser.add_argument(
        "--private-subnet-3-az",
        help="The availability zone for private subnet 3",
        default=None,
    )

    parser.add_argument(
        "--vpc-cidr",
        help="The CIDR for the VPC",
        default=None,
    )
    parser.add_argument(
        "--public-subnet-1-cidr",
        help="The CIDR for the public subnet",
        default=None,
    )
    parser.add_argument(
        "--private-subnet-1-cidr",
        help="The CIDR for private subnet 1",
        default=None,
    )
    parser.add_argument(
        "--private-subnet-2-cidr",
        help="The CIDR for private subnet 2",
        default=None,
    )
    parser.add_argument(
        "--private-subnet-3-cidr",
        help="The CIDR for private subnet 3",
        default=None,
    )

    # PostgresUrl
    parser.add_argument(
        "--managed-postgres",
        help="Spin up an RDS instance to use as the datastore",
        default=None,
        choices=[None, "true", "false"],
    )
    parser.add_argument(
        "--encrypt-database",
        help="Whether to encrypt the database",
        default="false",
        choices=[None, "true", "false"],
    )
    parser.add_argument(
        "--postgres-alternative-host",
        help="Use an external host for postgres (but the same secrets)",
        default=None,
    )

    # Clickhouse
    parser.add_argument(
        "--managed-clickhouse",
        help="Spin up a Clickhouse Instance for faster analytics",
        default=None,
        choices=[None, "true", "false"],
    )
    parser.add_argument(
        "--clickhouse-instance-type",
        help="The instance type for the Clickhouse instance",
        default=None,
    )

    # ElastiCacheClusterId
    parser.add_argument("--elasticache-cluster-host", help="The ElastiCacheCluster host to use", default=None)
    parser.add_argument(
        "--elasticache-cluster-port", help="The ElastiCacheCluster host to use", default=None, type=int
    )

    # SecurityGroupId, SubnetIds
    parser.add_argument("--security-group-id", help="The security group ID to use", default=None)
    parser.add_argument("--subnet-ids", help="The subnet IDs to use", default=None)

    # Advancd use only
    parser.add_argument(
        "--postgres-url",
        help="[Advanced] The postgres URL to use (if you are connecting to another VPC)",
        default=None,
    )
    parser.add_argument("--clickhouse-pg-url", help="[Advanced] The clickhouse PG URL to use", default=None)
    parser.add_argument("--clickhouse-connect-url", help="[Advanced] The clickhouse connect URL to use", default=None)
    parser.add_argument(
        "--clickhouse-catchup-etl-arn", help="[Advanced] The clickhouse catchup ETL ARN to use", default=None
    )

    parser.set_defaults(func=main)


def main(args):
    template = args.template or LATEST_TEMPLATE

    status = None
    try:
        statuses = cloudformation.describe_stacks(StackName=args.name)["Stacks"]
        if len(statuses) == 1:
            status = statuses[0]
        _logger.debug(status)
    except ClientError as e:
        if "does not exist" not in str(e):
            raise

    vpc_connect = args.vpc_connect
    if status and not vpc_connect:
        vpc_connect = "SecurityGroupId" in set(x["ParameterKey"] for x in status["Parameters"])

    if vpc_connect:
        PARAMS["SecurityGroupId"] = "security_group_id"
        PARAMS["SubnetIds"] = "subnet_ids"
        PARAMS["ElastiCacheClusterHost"] = "elasticache_cluster_host"
        PARAMS["ElastiCacheClusterPort"] = "elasticache_cluster_port"
        PARAMS["PostgresUrl"] = "postgres_url"
        PARAMS["ClickhouseCatchupEtlArn"] = "clickhouse_catchup_etl_arn"
        PARAMS["ClickhouseConnectUrl"] = "clickhouse_connect_url"
        PARAMS["ClickhousePGUrl"] = "clickhouse_pg_url"

        if args.template is None:
            template = "https://braintrust-cf.s3.amazonaws.com/braintrust-latest-vpc.yaml"

    exists = status is not None
    if exists and args.create:
        _logger.error(
            textwrap.dedent(
                f"""\
            Stack with name {args.name} already exists. Either delete it in the AWS console or
            remove the --create flag."""
            )
        )
        exit(1)
    elif not exists and not args.create:
        _logger.error(
            textwrap.dedent(
                f"""\
            Stack with name {args.name} does not exist. Either create it manually by following
            https://www.braintrustdata.com/docs/getting-started/install or use the --create flag."""
            )
        )
        exit(1)

    if not exists:
        _logger.info(f"Creating stack with name {args.name}")

        params = [
            {
                "ParameterKey": k,
                "ParameterValue": str(v),
            }
            for (k, v) in [
                (param, args.__dict__[arg_name] or DEFAULTS.get(param, None)) for (param, arg_name) in PARAMS.items()
            ]
            if v is not None
        ]
        _logger.info("Using params:", params)

        cloudformation.create_stack(
            StackName=args.name,
            TemplateURL=template,
            Parameters=params,
            Capabilities=CAPABILITIES,
        )

        for _ in range(120):
            status = cloudformation.describe_stacks(StackName=args.name)["Stacks"][0]
            if status["StackStatus"] != "CREATE_IN_PROGRESS":
                exists = True
                break
            _logger.info("Waiting for stack to be created...")
            time.sleep(5)
        else:
            _logger.error(
                textwrap.dedent(
                    """\
                Stack creation timed out. Please check the AWS console to see its status. You can also
                re-run this command without --create to continue the setup process once it's done."""
                )
            )
            exit(1)
        _logger.info(f"Stack with name {args.name} has been created with status: {status['StackStatus']}")
        exit(0)

    _logger.info(f"Stack with name {args.name} has status: {status['StackStatus']}")

    if not ("_COMPLETE" in status["StackStatus"] or "_FAILED" in status["StackStatus"]):
        _logger.info(f"Please re-run this command once the stack has finished creating or updating")
        exit(0)

    # Update params that have changed
    param_updates = {}
    for param, arg_name in PARAMS.items():
        if args.__dict__[arg_name] is not None:
            param_updates[param] = args.__dict__[arg_name]
    if len(param_updates) > 0 or args.update_template:
        template_kwargs = {"TemplateURL": template} if args.update_template else {"UsePreviousTemplate": True}
        _logger.info(
            f"Updating stack with name {args.name} with params: {param_updates} and template: {template_kwargs}"
        )

        stack = cloudformation.describe_stacks(StackName=args.name)["Stacks"][0]
        cloudformation.update_stack(
            StackName=args.name,
            Parameters=[
                {"ParameterKey": param, "ParameterValue": str(update)} for (param, update) in param_updates.items()
            ]
            + [
                {"ParameterKey": param["ParameterKey"], "UsePreviousValue": True}
                for param in stack["Parameters"]
                if param["ParameterKey"] not in param_updates and param["ParameterKey"] not in REMOVED_PARAMS
            ],
            Capabilities=CAPABILITIES,
            **template_kwargs,
        )

        for _ in range(120):
            status = cloudformation.describe_stacks(StackName=args.name)["Stacks"][0]
            if status["StackStatus"] != "UPDATE_IN_PROGRESS":
                exists = True
                break
            _logger.info("Waiting for stack to be updated...")
            time.sleep(5)
        else:
            _logger.error(
                textwrap.dedent(
                    """\
                Stack update timed out. Please check the AWS console to see its status. You can also
                re-run this command to try again."""
                )
            )
            exit(1)

        function_url = [x for x in status["Outputs"] if x["OutputKey"] == "EndpointURL"]
        if function_url:
            function_url = function_url[0]["OutputValue"]
        else:
            function_url = None
        _logger.info(f"Stack with name {args.name} has been updated with status: {status['StackStatus']}")
        _logger.info(f"Endpoint URL: {function_url}")



import logging
import os
import re
import subprocess
import threading
from functools import lru_cache as _cache
from typing import List, Optional

from braintrust_core.git_fields import GitMetadataSettings, RepoInfo

# https://stackoverflow.com/questions/48399498/git-executable-not-found-in-python
os.environ["GIT_PYTHON_REFRESH"] = "quiet"
try:
    import git
except ImportError:
    git = None

_logger = logging.getLogger("braintrust.gitutil")
_gitlock = threading.RLock()


@_cache(1)
def _current_repo():
    if git is None:
        # If the git module is not available, we can't do anything.
        return None

    try:
        return git.Repo(search_parent_directories=True)
    except git.exc.InvalidGitRepositoryError:
        return None


@_cache(1)
def _get_base_branch(remote=None):
    repo = _current_repo()
    remote = repo.remote(**({} if remote is None else {"name": remote})).name

    # NOTE: This should potentially be configuration that we derive from the project,
    # instead of spending a second or two computing it each time we run an experiment.

    # To speed this up in the short term, we pick from a list of common names
    # and only fall back to the remote origin if required.
    COMMON_BASE_BRANCHES = ["main", "master", "develop"]
    repo_branches = set(b.name for b in repo.branches)
    if sum(b in repo_branches for b in COMMON_BASE_BRANCHES) == 1:
        for b in COMMON_BASE_BRANCHES:
            if b in repo_branches:
                return (remote, b)
        raise RuntimeError("Impossible")

    try:
        s = subprocess.check_output(["git", "remote", "show", "origin"]).decode()
        match = re.search(r"\s*HEAD branch:\s*(.*)$", s, re.MULTILINE)
        if match is None:
            raise RuntimeError("Could not find HEAD branch in remote " + remote)
        branch = match.group(1)
    except Exception as e:
        _logger.warning(f"Could not find base branch for remote {remote}", e)
        branch = "main"
    return (remote, branch)


def _get_base_branch_ancestor(remote=None):
    try:
        remote_name, base_branch = _get_base_branch(remote)
    except Exception as e:
        _logger.warning(
            f"Skipping git metadata. This is likely because the repository has not been published to a remote yet. {e}"
        )
        return None

    head = "HEAD" if _current_repo().is_dirty() else "HEAD^"
    try:
        return subprocess.check_output(["git", "merge-base", head, f"{remote_name}/{base_branch}"]).decode().strip()
    except subprocess.CalledProcessError as e:
        # _logger.warning(f"Could not find a common ancestor with {remote_name}/{base_branch}")
        return None


def get_past_n_ancestors(n=10, remote=None):
    with _gitlock:
        repo = _current_repo()
        if repo is None:
            return

        ancestor_output = _get_base_branch_ancestor()
        if ancestor_output is None:
            return
        ancestor = repo.commit(ancestor_output)
        for _ in range(n):
            yield ancestor.hexsha
            try:
                if ancestor.parents:
                    ancestor = ancestor.parents[0]
                else:
                    break
            except ValueError:
                # Since parents are fetched on-demand, this can happen if the
                # downloaded repo does not have information for this commit's
                # parent.
                break


def attempt(op):
    try:
        return op()
    except (TypeError, ValueError, git.GitCommandError):
        return None


def truncate_to_byte_limit(input_string, byte_limit=65536):
    encoded = input_string.encode("utf-8")
    if len(encoded) <= byte_limit:
        return input_string
    return encoded[:byte_limit].decode("utf-8", errors="ignore")


def get_repo_info(settings: Optional[GitMetadataSettings] = None):
    if settings is None:
        settings = GitMetadataSettings()

    if settings.collect == "none":
        return None

    repo = repo_info()
    if repo is None or settings.collect == "all":
        return repo

    return RepoInfo(**{k: v if k in settings.fields else None for k, v in repo.as_dict().items()})


def repo_info():
    with _gitlock:
        repo = _current_repo()
        if repo is None:
            return None

        commit = None
        commit_message = None
        commit_time = None
        author_name = None
        author_email = None
        tag = None
        branch = None
        git_diff = None

        dirty = repo.is_dirty()

        commit = attempt(lambda: repo.head.commit.hexsha.strip())
        commit_message = attempt(lambda: repo.head.commit.message.strip())
        commit_time = attempt(lambda: repo.head.commit.committed_datetime.isoformat())
        author_name = attempt(lambda: repo.head.commit.author.name.strip())
        author_email = attempt(lambda: repo.head.commit.author.email.strip())
        tag = attempt(lambda: repo.git.describe("--tags", "--exact-match", "--always"))

        branch = attempt(lambda: repo.active_branch.name)

        if dirty:
            git_diff = attempt(lambda: truncate_to_byte_limit(repo.git.diff("HEAD")))

        return RepoInfo(
            commit=commit,
            branch=branch,
            tag=tag,
            dirty=dirty,
            author_name=author_name,
            author_email=author_email,
            commit_message=commit_message,
            commit_time=commit_time,
            git_diff=git_diff,
        )



import logging
import time
from concurrent.futures import ThreadPoolExecutor

from ...aws import cloudformation, logs

_logger = logging.getLogger("braintrust.install.logs")


def build_parser(subparsers, parents):
    parser = subparsers.add_parser("logs", help="Capture recent logs", parents=parents)
    parser.add_argument("name", help="Name of the CloudFormation stack to collect logs from")
    parser.add_argument("--service", help="Name of the service", default="api", choices=["api"])
    parser.add_argument("--hours", help="Number of seconds in the past to collect logs from", default=1, type=float)
    parser.set_defaults(func=main)


def main(args):
    stacks = cloudformation.describe_stacks(StackName=args.name)["Stacks"]
    if len(stacks) == 0:
        raise ValueError(f"Stack with name {args.name} does not exist")
    if len(stacks) > 1:
        raise ValueError(f"Multiple stacks with name {args.name} exist")
    stack = stacks[0]
    _logger.debug(stack)

    log_group_name = None
    if args.service == "api":
        lambda_function = [x for x in stack["Outputs"] if x["OutputKey"] == "APIHandlerName"]
        if len(lambda_function) != 1:
            raise ValueError(f"Expected 1 APIHandlerName, found {len(lambda_function)} ({lambda_function}))")
        log_group_name = f"/aws/lambda/{lambda_function[0]['OutputValue']}"
    else:
        raise ValueError(f"Unknown service {args.service}")

    start_time = int(time.time() - 3600 * args.hours) * 1000

    all_streams = []
    first_start_time = None
    nextToken = None

    while first_start_time is None or first_start_time >= start_time:
        kwargs = {}
        if nextToken is not None:
            kwargs["nextToken"] = nextToken

        stream_resp = logs.describe_log_streams(logGroupName=log_group_name, descending=True, **kwargs)

        first_start_time = min(s["firstEventTimestamp"] for s in stream_resp["logStreams"])
        nextToken = stream_resp.get("nextToken")

        streams = [s for s in stream_resp["logStreams"] if s["firstEventTimestamp"] >= start_time]
        streams.sort(key=lambda x: x["firstEventTimestamp"])
        all_streams = streams + all_streams

    _logger.debug(all_streams)

    def get_events(stream):
        return logs.get_log_events(
            logGroupName=log_group_name,
            logStreamName=stream["logStreamName"],
            startTime=start_time,
            startFromHead=True,
        )

    with ThreadPoolExecutor(8) as executor:
        events = executor.map(get_events, all_streams)

    last_ts = None
    for stream, log in zip(all_streams, events):
        print(f"---- {stream['logStreamName']}")
        for event in log["events"]:
            print(event)



import time

from braintrust_core.span_types import SpanTypeAttribute
from braintrust_core.util import merge_dicts

from .logger import start_span


class NamedWrapper:
    def __init__(self, wrapped):
        self.__wrapped = wrapped

    def __getattr__(self, name):
        return getattr(self.__wrapped, name)


def postprocess_streaming_results(all_results):
    role = None
    content = None
    tool_calls = None
    finish_reason = None
    for result in all_results:
        delta = result["choices"][0]["delta"]
        if role is None and delta.get("role") is not None:
            role = delta.get("role")

        if delta.get("finish_reason") is not None:
            finish_reason = delta.get("finish_reason")

        if delta.get("content") is not None:
            content = (content or "") + delta.get("content")
        if delta.get("tool_calls") is not None:
            if tool_calls is None:
                tool_calls = [
                    {
                        "id": delta["tool_calls"][0]["id"],
                        "type": delta["tool_calls"][0]["type"],
                        "function": delta["tool_calls"][0]["function"],
                    }
                ]
            else:
                tool_calls[0]["function"]["arguments"] += delta["tool_calls"][0]["function"]["arguments"]

    return [
        {
            "index": 0,
            "message": {
                "role": role,
                "content": content,
                "tool_calls": tool_calls,
            },
            "logprobs": None,
            "finish_reason": finish_reason,
        }
    ]


class ChatCompletionWrapper:
    def __init__(self, create_fn, acreate_fn):
        self.create_fn = create_fn
        self.acreate_fn = acreate_fn

    def create(self, *args, **kwargs):
        params = self._parse_params(kwargs)
        stream = kwargs.get("stream", False)

        span = start_span(name="OpenAI Chat Completion", span_attributes={"type": SpanTypeAttribute.LLM}, **params)
        should_end = True

        try:
            start = time.time()
            raw_response = self.create_fn(*args, **kwargs)
            if stream:

                def gen():
                    try:
                        first = True
                        all_results = []
                        for item in raw_response:
                            if first:
                                span.log(
                                    metrics={
                                        "time_to_first_token": time.time() - start,
                                    }
                                )
                                first = False
                            all_results.append(item if isinstance(item, dict) else item.dict())
                            yield item

                        span.log(output=postprocess_streaming_results(all_results))
                    finally:
                        span.end()

                should_end = False
                return gen()
            else:
                log_response = raw_response if isinstance(raw_response, dict) else raw_response.dict()
                span.log(
                    metrics={
                        "time_to_first_token": time.time() - start,
                        "tokens": log_response["usage"]["total_tokens"],
                        "prompt_tokens": log_response["usage"]["prompt_tokens"],
                        "completion_tokens": log_response["usage"]["completion_tokens"],
                    },
                    output=log_response["choices"],
                )
                return raw_response
        finally:
            if should_end:
                span.end()

    async def acreate(self, *args, **kwargs):
        params = self._parse_params(kwargs)
        stream = kwargs.get("stream", False)

        span = start_span(name="OpenAI Chat Completion", span_attributes={"type": SpanTypeAttribute.LLM}, **params)
        should_end = True

        try:
            start = time.time()
            raw_response = await self.acreate_fn(*args, **kwargs)
            if stream:

                async def gen():
                    try:
                        first = True
                        all_results = []
                        async for item in raw_response:
                            if first:
                                span.log(
                                    metrics={
                                        "time_to_first_token": time.time() - start,
                                    }
                                )
                                first = False
                            all_results.append(item if isinstance(item, dict) else item.dict())
                            yield item

                        span.log(output=postprocess_streaming_results(all_results))
                    finally:
                        span.end()

                should_end = False
                return gen()
            else:
                log_response = raw_response if isinstance(raw_response, dict) else raw_response.dict()
                span.log(
                    metrics={
                        "tokens": log_response["usage"]["total_tokens"],
                        "prompt_tokens": log_response["usage"]["prompt_tokens"],
                        "completion_tokens": log_response["usage"]["completion_tokens"],
                    },
                    output=log_response["choices"],
                )
                return raw_response
        finally:
            if should_end:
                span.end()

    @classmethod
    def _parse_params(cls, params):
        # First, destructively remove span_info
        ret = params.pop("span_info", {})

        # Then, copy the rest of the params
        params = {**params}
        messages = params.pop("messages", None)
        merge_dicts(
            ret,
            {
                "input": messages,
                "metadata": params,
            },
        )

        return ret


class EmbeddingWrapper:
    def __init__(self, create_fn, acreate_fn):
        self.create_fn = create_fn
        self.acreate_fn = acreate_fn

    def create(self, *args, **kwargs):
        params = self._parse_params(kwargs)

        with start_span(name="OpenAI Embedding", span_attributes={"type": SpanTypeAttribute.LLM}, **params) as span:
            raw_response = self.create_fn(*args, **kwargs)
            log_response = raw_response if isinstance(raw_response, dict) else raw_response.dict()
            span.log(
                metrics={
                    "tokens": log_response["usage"]["total_tokens"],
                    "prompt_tokens": log_response["usage"]["prompt_tokens"],
                },
                # TODO: Add a flag to control whether to log the full embedding vector,
                # possibly w/ JSON compression.
                output={"embedding_length": len(log_response["data"][0]["embedding"])},
            )
            return raw_response

    async def acreate(self, *args, **kwargs):
        params = self._parse_params(kwargs)

        with start_span(name="OpenAI Embedding", span_attributes={"type": SpanTypeAttribute.LLM}, **params) as span:
            raw_response = await self.acreate_fn(*args, **kwargs)
            log_response = raw_response if isinstance(raw_response, dict) else raw_response.dict()
            span.log(
                metrics={
                    "tokens": log_response["usage"]["total_tokens"],
                    "prompt_tokens": log_response["usage"]["prompt_tokens"],
                },
                # TODO: Add a flag to control whether to log the full embedding vector,
                # possibly w/ JSON compression.
                output={"embedding_length": len(log_response["data"][0]["embedding"])},
            )
            return raw_response

    @classmethod
    def _parse_params(cls, params):
        # First, destructively remove span_info
        ret = params.pop("span_info", {})

        params = {**params}
        input = params.pop("input", None)

        merge_dicts(
            ret,
            {
                "input": input,
                "metadata": params,
            },
        )

        return ret


class ChatCompletionV0Wrapper(NamedWrapper):
    def __init__(self, chat):
        self.__chat = chat
        super().__init__(chat)

    def create(self, *args, **kwargs):
        return ChatCompletionWrapper(self.__chat.create, self.__chat.acreate).create(*args, **kwargs)

    async def acreate(self, *args, **kwargs):
        return await ChatCompletionWrapper(self.__chat.create, self.__chat.acreate).acreate(*args, **kwargs)


class EmbeddingV0Wrapper(NamedWrapper):
    def __init__(self, embedding):
        self.__embedding = embedding
        super().__init__(embedding)

    def create(self, *args, **kwargs):
        return EmbeddingWrapper(self.__embedding.create, self.__embedding.acreate).create(*args, **kwargs)

    async def acreate(self, *args, **kwargs):
        return await ChatCompletionWrapper(self.__embedding.create, self.__embedding.acreate).acreate(*args, **kwargs)


# This wraps 0.*.* versions of the openai module, eg https://github.com/openai/openai-python/tree/v0.28.1
class OpenAIV0Wrapper(NamedWrapper):
    def __init__(self, openai):
        super().__init__(openai)
        self.ChatCompletion = ChatCompletionV0Wrapper(openai.ChatCompletion)
        self.Embedding = EmbeddingV0Wrapper(openai.Embedding)


class CompletionsV1Wrapper(NamedWrapper):
    def __init__(self, completions):
        self.__completions = completions
        super().__init__(completions)

    def create(self, *args, **kwargs):
        return ChatCompletionWrapper(self.__completions.create, None).create(*args, **kwargs)


class EmbeddingV1Wrapper(NamedWrapper):
    def __init__(self, embedding):
        self.__embedding = embedding
        super().__init__(embedding)

    def create(self, *args, **kwargs):
        return EmbeddingWrapper(self.__embedding.create, None).create(*args, **kwargs)


class AsyncCompletionsV1Wrapper(NamedWrapper):
    def __init__(self, completions):
        self.__completions = completions
        super().__init__(completions)

    async def create(self, *args, **kwargs):
        return await ChatCompletionWrapper(None, self.__completions.create).acreate(*args, **kwargs)


class AsyncEmbeddingV1Wrapper(NamedWrapper):
    def __init__(self, embedding):
        self.__embedding = embedding
        super().__init__(embedding)

    async def create(self, *args, **kwargs):
        return await EmbeddingWrapper(None, self.__embedding.create).acreate(*args, **kwargs)


class ChatV1Wrapper(NamedWrapper):
    def __init__(self, chat):
        super().__init__(chat)

        import openai

        if type(chat.completions) == openai.resources.chat.completions.AsyncCompletions:
            self.completions = AsyncCompletionsV1Wrapper(chat.completions)
        else:
            self.completions = CompletionsV1Wrapper(chat.completions)


# This wraps 1.*.* versions of the openai module, eg https://github.com/openai/openai-python/tree/v1.1.0
class OpenAIV1Wrapper(NamedWrapper):
    def __init__(self, openai):
        super().__init__(openai)
        import openai as oai

        self.chat = ChatV1Wrapper(openai.chat)

        if type(openai.embeddings) == oai.resources.embeddings.AsyncEmbeddings:
            self.embeddings = AsyncEmbeddingV1Wrapper(openai.embeddings)
        else:
            self.embeddings = EmbeddingV1Wrapper(openai.embeddings)


def wrap_openai(openai):
    """
    Wrap the openai module (pre v1) or OpenAI instance (post v1) to add tracing.
    If Braintrust is not configured, this is a no-op.

    :param openai: The openai module or OpenAI object
    """
    if hasattr(openai, "chat") and hasattr(openai.chat, "completions"):
        return OpenAIV1Wrapper(openai)
    else:
        return OpenAIV0Wrapper(openai)
